/* thermo.c -- Moist thermodynamics package -- derived from diabat2.
 * Variable naming convection:
 * th: Potential temperature (K)
 * pi: Exner function (J/kg/K)
 * td: Dewpoint temperature (K)
 * ent: Specific moist entropy (J/kg/K)
 * rt: Total cloud water (vapor plus droplets) mixing ratio (g/g)
 * rsat: Saturation mixing ratio (g/g)
 * p: Pressure (pa)
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "thermo.h"

static double thetasat(double ent,double pi,double thstart,int nloops);

/* rs - compute saturation mixing ratio (g/g) as function of
 * potential temp (K) and Exner function (J/kg/K)
 * use simple formula
 */
double rs(double th,double pi)
{
  double p,es,rsat;
  p = P_REF*pow(pi/CP,1./KAPPA);
  es = escalc(th*pi/CP);
  rsat = EPSILON*es/p;
  return(rsat);
}

/* escalc -- compute vapor pressure (pa) from dewpoint (K) -- Teten formula */
double escalc(double td)
{
  return(611.2*exp(17.67*(td - 273.15)/(td - 29.65)));
}

/* thetacomp2 -- do this computation by a lookup table */
#define IENTMAX 501
#define IPIMAX 451
#define DENT 1.
#define DPI 2.
#define ENT0 0.
#define PI0 200.

double thetacomp2(double ent,double rt,double pi)
{
  static double v[IENTMAX][IPIMAX];
  static int begin = 1;
  int ient,ipi,jent,jpi;
  double entval,pival,aent,api,oent,opi,theta;

  /* fill the lookup table the first time around */
  if (begin) {
    for (ient = 0; ient < IENTMAX; ient++) {
      for (ipi = 0; ipi < IPIMAX; ipi++) {
	entval = ENT0 + ient*DENT;
	pival = PI0 + ipi*DPI;
	v[ient][ipi] = thetasat(entval,pival,TREF*exp(ent/CP),200);
      }
    }
    begin = 0;
  }
  
  /* compute test value of theta */
  theta = TREF*exp(ent/CP - ALFA*rt);
  
  /* if unsaturated, don't do anything */
  if (rt < rs(theta,pi)) {
  }
  
  /* if saturated, we interpolate */
  else {
    ient = (ent - ENT0)/DENT;
    if (ient < 0 || ient > IENTMAX - 2) {
      fprintf(stderr,"thermo: ent = %g is out of lookup table range\n",ent);
      exit(1);
    }
    jent = ient + 1;
    ipi = (pi - PI0)/DPI;
    if (ipi < 0 || ipi > IPIMAX - 2) {
      fprintf(stderr,"thermo: pi = %g is out of lookup table range\n",pi);
      exit(1);
    }
    jpi = ipi + 1;
    aent = (ent - (ENT0 + DENT*ient))/DENT;
    oent = 1. - aent;
    api = (pi - (PI0 + DPI*ipi))/DPI;
    opi = 1. - api;
    theta = oent*(opi*v[ient][ipi] + api*v[ient][jpi])
      + aent*(opi*v[jent][ipi] + api*v[jent][jpi]);
  }
  return(theta);
}

/* thetacomp -- compute potential temperature from equivalent
 * potential temperature, total water mixing ratio, and Exner function.
 */
double thetacomp(double ent,double rt,double pi)
{
  double theta;
  
  /* compute test value of theta */
  theta = TREF*exp(ent/CP - ALFA*rt);
  
  /* if unsaturated, don't do anything */
  if (rt < rs(theta,pi)) {
  }
  
  /* if saturated -- we have to do a Newton's method iteration */
  else {
    theta = thetasat(ent,pi,TREF*exp(ent/CP),200);
  }
  
  /* return with theta value */
  return(theta);
}

/* thetasat -- assume saturation and calculate theta -- really
 * theta-l */
static double thetasat(double ent,double pi,double thstart,int nloops)
{
  double theta,rsval,drsdth,dtheta;
  int loop;
  theta = thstart;
  for (loop = 0; loop < nloops; loop++) {
    rsval = rs(theta,pi);
    drsdth = rs(theta + 1.,pi) - rsval;
    dtheta = - (log(theta/TREF) + ALFA*rsval - ent/CP)/
      (1./theta + ALFA*drsdth);
    if (dtheta > 10.) dtheta = 10.;
    if (dtheta < -10.) dtheta = -10.;
    theta += dtheta;
    if (fabs(dtheta) < 0.001) return(theta);
  }
  fprintf(stderr,"thermo: no thetasat conv: ent = %g; pi = %g\n",ent,pi);
  exit(1);
}

/* calculate the equivalent potential temperature from the entropy */
double thecomp(double ent)
{
  return(TREF*exp(ent/CP));
}

/* calculate the entropy source term from the thetae source term */
double entrate(double therate,double the)
{
  return(CP*therate/the);
}

/* calculate entropy */
double entcomp(double th,double rt,double pi)
{
  return(entcomp2(th,rt,rs(th,pi)));
}

/* calculate entropy given the saturation mixing ratio */
double entcomp2(double th,double rt,double rsat)
{
  double rvap;
  rvap = rt > rsat ? rsat : rt;
  return(CP*(log(th/TREF) + ALFA*rvap));
}

/* entscomp -- compute saturate entropy value given theta (K) and
 * pi (J/kg/K)
 */
double entscomp(double th,double pi)
{
  return(CP*(log(th/TREF) + ALFA*rs(th,pi)));
}

/* gmoist -- compute (d theta/dz) for moist adiabatic parcel */
double gmoist(double th,double pi)
{
  double drsdpi,drsdth,rate;
  drsdpi = (rs(th,pi + 0.05) - rs(th,pi - 0.05))/0.1;
  drsdth = (rs(th + 0.05,pi) - rs(th - 0.05,pi))/0.1;
  rate = GEE*ALFA*drsdpi/(1. + th*ALFA*drsdth);
  return(rate);
}

/* exner -- compute Exner function from pressure */
double exner(double p)
{
  return(CP*pow(p/P_REF,KAPPA));
}
