/* timestep.c -- This routine time steps the model by one step using
 * the Lax-Wendroff method.  The variables are updated in place.
 */ 

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include "thermo.h"
#include "indexing.h"
#include "doublebuff.h"

/* NU is the adjustment rate to keep surface pressure at a reference
 * value */
#define NU 0.001

/* MPIFIELDS is the number of fields transferred between processes */
#define MPIFIELDS 10

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* compute the rank of a process from the position of its domain */
int getrank(int gx, int igx, int igy)
{
  return(igx + gx*(igy - 1));
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* transfer -- transfer data from one process to another
 * buffin: buffer containing stuff sent from the source process
 * buffout: buffer containing stuff received by the destination process
 * getfrom: rank of the source process
 * sendto: rank of the destination process
 * size: size of the data in the buffers
 * odd_flag: true if source process has an odd value of igx or igy
 * mycom: the MPI communicator
 */
void transfer(double *buffin, double *buffout, int getfrom, int sendto,
	      long size, int odd_flag, MPI_Comm mycom)
{
  MPI_Status status;

  /* odd to even transfers */
  if (odd_flag) {   /* set up send */
    MPI_Send(buffin,size,MPI_DOUBLE,sendto,1,mycom);
  }
  else {   /* set up receive */
    MPI_Recv(buffout,size,MPI_DOUBLE,getfrom,1,mycom,&status);
  }
  
  /* even to odd transfers */
  if (!odd_flag) {   /* set up send */
    MPI_Send(buffin,size,MPI_DOUBLE,sendto,1,mycom);
  }
  else {   /* set up receive */
    MPI_Recv(buffout,size,MPI_DOUBLE,getfrom,1,mycom,&status);
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* timestep: */

void timestep(int mpisize, int gx, int gy, MPI_Comm mycom, int rank,
	      int igx, int igy, int nx, int nxs, int nx1, int ny, int nys,
	      int ny1, int nz, double dx, double dy, double dz, double dt,
	      double eddy, double filter, double lambda, double drh,
	      double crain, double csnow, double cevap, double wtermw,
	      double wtermi, double cld, double cdrag, double ueffmin,
	      double *sentwtg, double *srtwtg, double *svx, double *svy,
	      double *mu, double *vx0, double *vy0, double psmean,
	      double *rho, double *vx, double *vy, double *vz, double *p,
	      double *th, double *ent, double *rt, double *r, double *rc,
	      double *rr, double *kk, double *kv, double *vxcuflux,
	      double *vxcufluxcum, double*vycuflux, double *vycufluxcum,
	      double *entcuflux, double *entcufluxcum, double *rtcuflux,
	      double *rtcufluxcum, double *rrcuflux, double *rrcufluxcum,
	      double *sentrad, double *sentradcum, double *srtrain,
	      double *srtraincum, double *entflux, double *entfluxcum,
	      double *rtflux, double *rtfluxcum, double*vxflux,
	      double *vxfluxcum, double *vyflux, double *vyfluxcum,
	      double *rainflux, double *rainfluxcum, double *stemp)
{
  static int firstcall = 1;
  int ix, iy, iz, oddigx, oddigy, igxl, igxr, igyb, igyt;
  int rankl, rankr, rankb, rankt;
  long bsizex,bsizey, position;
  static double *bbuffinx, *bbuffoutx,*bbuffiny,*bbuffouty;
  long jsize, lsize, i, j, l, k, j0, j1, j2, jl, jm, jn, l0, ln, ilo;
  double dthalf, dzhalf, crho, sth, oldrho, dxx, dyy, dzz, dxy, dyz,
    dzx, dsq, bvsq, kcriterion, rsat, rh, zgrad, dstab, mstab,
    newrain, entsurf, rtsurf, pisurf, thsurf, ueff, entflux1,
    rtflux1, vxflux1, vyflux1, cprecip, temperature;
  static double *fruu, *fruv, *fruw, *frvv, *frvw, *frww, *fru, *frv,
    *frw, *fruent, *frvent, *frwent, *frurt, *frvrt, *frwrt,
    *frurr, *frvrr, *frwrr;
  static double *sfruu, *sfruv, *sfruw, *sfrvv, *sfrvw, *sfrww, *sfru,
    *sfrv, *sfrw, *sfruent, *sfrvent, *sfrwent, *sfrurt, *sfrvrt,
    *sfrwrt, *sfrurr, *sfrvrr, *sfrwrr, *skk, *skv;
  static double *xp, *sp, *xrho, *srho, *xrent, *srent, *xrrt, *srrt,
    *xrrr, *srrr, *xvx, *xvy, *xvz, *xth, *xent, *xrt, *xrr, *xwterm,
    *swterm, *xthe;

  /* set up stuff for transfers */
  if (mpisize > 0) {
    if (getrank(gx,igx,igy) != rank) {
      fprintf(stderr,"rank %d != getrank(%d,%d,%d) %d\n",rank,gx,igx,igy,
	      getrank(gx,igx,igy));
      exit(50);
    }
    igxl = igx - 1;
    if (igxl < 1) igxl = gx;
    igxr = igx + 1;
    if (igxr > gx) igxr = 1;
    igyb = igy - 1;
    if (igyb < 1) igyb = gy;
    igyt = igy + 1;
    if (igyt > gy) igyt = 1;
    oddigx = igx & 1;
    oddigy = igy & 1;
    rankl = getrank(gx,igxl,igy);
    rankr = getrank(gx,igxr,igy);
    rankb = getrank(gx,igx,igyb);
    rankt = getrank(gx,igx,igyt);
  }
  else {
    oddigx = 0;
    oddigy = 0;
    rankl = 0;
    rankr = 0;
    rankb = 0;
    rankt = 0;
  }

  /* allocate space */
  jsize = (nxs + 1)*(nys + 1)*(nz + 1);
  lsize = (nxs + 2)*(nys + 2)*nz;
  bsizex = MPIFIELDS*(nys + 1)*nz;
  bsizey = MPIFIELDS*(nxs + 3)*nz;
  if (firstcall) {
    firstcall = 0;
    fruu = doublebuff(lsize);
    fruv = doublebuff(lsize);
    fruw = doublebuff(lsize);
    frvv = doublebuff(lsize);
    frvw = doublebuff(lsize);
    frww = doublebuff(lsize);
    fru = doublebuff(lsize);
    frv = doublebuff(lsize);
    frw = doublebuff(lsize);
    fruent = doublebuff(lsize);
    frvent = doublebuff(lsize);
    frwent = doublebuff(lsize);
    frurt = doublebuff(lsize);
    frvrt = doublebuff(lsize);
    frwrt = doublebuff(lsize);
    frurr = doublebuff(lsize);
    frvrr = doublebuff(lsize);
    frwrr = doublebuff(lsize);
    sfruu = doublebuff(jsize);
    sfruv = doublebuff(jsize);
    sfruw = doublebuff(jsize);
    sfrvv = doublebuff(jsize);
    sfrvw = doublebuff(jsize);
    sfrww = doublebuff(jsize);
    sfru = doublebuff(jsize);
    sfrv = doublebuff(jsize);
    sfrw = doublebuff(jsize);
    sfruent = doublebuff(jsize);
    sfrvent = doublebuff(jsize);
    sfrwent = doublebuff(jsize);
    sfrurt = doublebuff(jsize);
    sfrvrt = doublebuff(jsize);
    sfrwrt = doublebuff(jsize);
    sfrurr = doublebuff(jsize);
    sfrvrr = doublebuff(jsize);
    sfrwrr = doublebuff(jsize);
    srho = doublebuff(jsize);
    xrho = doublebuff(lsize);
    swterm = doublebuff(jsize);
    xwterm = doublebuff(lsize);
    sp = doublebuff(jsize);
    xp = doublebuff(lsize);
    srent = doublebuff(jsize);
    xrent = doublebuff(lsize);
    srrt = doublebuff(jsize);
    xrrt = doublebuff(lsize);
    srrr = doublebuff(jsize);
    xrrr = doublebuff(lsize);
    xvx = doublebuff(lsize);
    xvy = doublebuff(lsize);
    xvz = doublebuff(lsize);
    xth = doublebuff(lsize);
    xent = doublebuff(lsize);
    xthe = doublebuff(lsize);
    xrt = doublebuff(lsize);
    xrr = doublebuff(lsize);
    skk = doublebuff(jsize);
    skv = doublebuff(jsize);
    bbuffinx = doublebuff(bsizex);
    bbuffoutx = doublebuff(bsizex);
    bbuffiny = doublebuff(bsizey);
    bbuffouty = doublebuff(bsizey);
  }

  /* preliminary stuff */
  dthalf = dt/2.;
  dzhalf = dz/2.;

  /* transfer various fields to expanded L grid */
  for (ix = 0; ix < nxs; ix++) {
    for (iy = 0; iy < nys; iy++) {
      for (iz = 0; iz < nz; iz++) {
	l = L(ix,iy,iz);
	i = I(ix,iy,iz);
	xp[l] = p[i];
	xrho[l] = rho[i];
	xvx[l] = vx[i];
	xvy[l] = vy[i];
	xvz[l] = vz[i];
	xth[l] = th[i];
	xent[l] = ent[i];
	xrt[l] = rt[i];
	xrr[l] = rr[i];

	/* compute the temperature and define the terminal velocity */
	temperature = th[i]*pow(p[i]/P_REF,KAPPA);
	xwterm[l] = temperature > FREEZING ? wtermw : wtermi;
      }
    }
  }

  /* extend fields in plus x direction periodically */
  if (gx == 1) {   /* no parallel processing in x direction */
    for (iy = 0; iy < nys; iy++) {
      for (iz = 0; iz < nz; iz++) {
	ln = L(nxs,iy,iz);
	l0 = L(0,iy,iz);
	xp[ln] = xp[l0];
	xrho[ln] = xrho[l0];
	xwterm[ln] = xwterm[l0];
	xvx[ln] = xvx[l0];
	xvy[ln] = xvy[l0];
	xvz[ln] = xvz[l0];
	xent[ln] = xent[l0];
	xrt[ln] = xrt[l0];
	xrr[ln] = xrr[l0];
	xth[ln] = xth[l0];
      }
    }
  }
  else {   /* parallel processing */

    /* transfer needed stuff to buffer */
    position = 0;
    for (iy = 0; iy < nys; iy++) {
      for (iz = 0; iz < nz; iz++) {
	l0 = L(0,iy,iz);
	bbuffinx[position++] = xp[l0];
	bbuffinx[position++] = xrho[l0];
	bbuffinx[position++] = xwterm[l0];
	bbuffinx[position++] = xvx[l0];
	bbuffinx[position++] = xvy[l0];
	bbuffinx[position++] = xvz[l0];
	bbuffinx[position++] = xent[l0];
	bbuffinx[position++] = xrt[l0];
	bbuffinx[position++] = xrr[l0];
	bbuffinx[position++] = xth[l0];
      }
    }
    if (position > bsizex) {
      fprintf(stderr,"timestep: bbuffinx buffer not big enough\n");
      exit(51);
    }

    /* transfer stuff in negative x direction */
    transfer(bbuffinx,bbuffoutx,rankr,rankl,position,oddigx,mycom);

    /* extract stuff from buffer */
    position = 0;
    for (iy = 0; iy < nys; iy++) {
      for (iz = 0; iz < nz; iz++) {
	ln = L(nxs,iy,iz);
	xp[ln] = bbuffoutx[position++];
	xrho[ln] = bbuffoutx[position++];
	xwterm[ln] = bbuffoutx[position++];
	xvx[ln] = bbuffoutx[position++];
	xvy[ln] = bbuffoutx[position++];
	xvz[ln] = bbuffoutx[position++];
	xent[ln] = bbuffoutx[position++];
	xrt[ln] = bbuffoutx[position++];
	xrr[ln] = bbuffoutx[position++];
	xth[ln] = bbuffoutx[position++];
      }
    }
  }

  /* extend fields in minus x direction periodically */
  if (gx == 1) {   /* no parallel processing */
    for (iy = 0; iy <= nys; iy++) {
      for (iz = 0; iz < nz; iz++) {
	ln = L(-1,iy,iz);
	l0 = L(nxs - 1,iy,iz);
	xp[ln] = xp[l0];
	xrho[ln] = xrho[l0];
	xwterm[ln] = xwterm[l0];
	xvx[ln] = xvx[l0];
	xvy[ln] = xvy[l0];
	xvz[ln] = xvz[l0];
	xent[ln] = xent[l0];
	xrt[ln] = xrt[l0];
	xrr[ln] = xrr[l0];
	xth[ln] = xth[l0];
      }
    }
  }
  else {   /* parallel processing */

    /* transfer needed stuff to buffer */
    position = 0;
    for (iy = 0; iy <= nys; iy++) {
      for (iz = 0; iz < nz; iz++) {
	l0 = L(nxs - 1,iy,iz);
	bbuffinx[position++] = xp[l0];
	bbuffinx[position++] = xrho[l0];
	bbuffinx[position++] = xwterm[l0];
	bbuffinx[position++] = xvx[l0];
	bbuffinx[position++] = xvy[l0];
	bbuffinx[position++] = xvz[l0];
	bbuffinx[position++] = xent[l0];
	bbuffinx[position++] = xrt[l0];
	bbuffinx[position++] = xrr[l0];
	bbuffinx[position++] = xth[l0];
      }
    }
    if (position > bsizex) {
      fprintf(stderr,"timestep: bbuffinx buffer not big enough\n");
      exit(52);
    }

    /* transfer stuff in positive x direction */
    transfer(bbuffinx,bbuffoutx,rankl,rankr,position,oddigx,mycom);

    /* extract stuff from buffer */
    position = 0;
    for (iy = 0; iy <= nys; iy++) {
      for (iz = 0; iz < nz; iz++) {
	ln = L(-1,iy,iz);
	xp[ln] = bbuffoutx[position++];
	xrho[ln] = bbuffoutx[position++];
	xwterm[ln] = bbuffoutx[position++];
	xvx[ln] = bbuffoutx[position++];
	xvy[ln] = bbuffoutx[position++];
	xvz[ln] = bbuffoutx[position++];
	xent[ln] = bbuffoutx[position++];
	xrt[ln] = bbuffoutx[position++];
	xrr[ln] = bbuffoutx[position++];
	xth[ln] = bbuffoutx[position++];
      }
    }
  }

  /* extend fields in plus y direction periodically */
  if (gy == 1) {   /* no parallel processing in y direction */
    for (ix = -1; ix <= nxs; ix++) {
      for (iz = 0; iz < nz; iz++) {
	ln = L(ix,nys,iz);
	l0 = L(ix,0,iz);
	xp[ln] = xp[l0];
	xrho[ln] = xrho[l0];
	xwterm[ln] = xwterm[l0];
	xvx[ln] = xvx[l0];
	xvy[ln] = xvy[l0];
	xvz[ln] = xvz[l0];
	xent[ln] = xent[l0];
	xrt[ln] = xrt[l0];
	xrr[ln] = xrr[l0];
	xth[ln] = xth[l0];
      }
    }
  }
  else {   /* parallel processing */

    /* transfer needed stuff to buffer */
    position = 0;
    for (ix = -1; ix <= nxs; ix++) {
      for (iz = 0; iz < nz; iz++) {
	l0 = L(ix,0,iz);
	bbuffiny[position++] = xp[l0];
	bbuffiny[position++] = xrho[l0];
	bbuffiny[position++] = xwterm[l0];
	bbuffiny[position++] = xvx[l0];
	bbuffiny[position++] = xvy[l0];
	bbuffiny[position++] = xvz[l0];
	bbuffiny[position++] = xent[l0];
	bbuffiny[position++] = xrt[l0];
	bbuffiny[position++] = xrr[l0];
	bbuffiny[position++] = xth[l0];
      }
    }
    if (position > bsizey) {
      fprintf(stderr,"timestep: bbuffiny buffer not big enough\n");
      exit(53);
    }

    /* transfer stuff in negative y direction */
    transfer(bbuffiny,bbuffouty,rankt,rankb,position,oddigy,mycom);

    /* extract stuff from buffer */
    position = 0;
    for (ix = -1; ix <= nxs; ix++) {
      for (iz = 0; iz < nz; iz++) {
	ln = L(ix,nys,iz);
	xp[ln] = bbuffouty[position++];
	xrho[ln] = bbuffouty[position++];
	xwterm[ln] = bbuffouty[position++];
	xvx[ln] = bbuffouty[position++];
	xvy[ln] = bbuffouty[position++];
	xvz[ln] = bbuffouty[position++];
	xent[ln] = bbuffouty[position++];
	xrt[ln] = bbuffouty[position++];
	xrr[ln] = bbuffouty[position++];
	xth[ln] = bbuffouty[position++];
      }
    }
  }

  /* extend fields in minus y direction periodically */
  if (gy == 1) {   /* no parallel processing */
    for (ix = -1; ix <= nxs; ix++) {
      for (iz = 0; iz < nz; iz++) {
	ln = L(ix,-1,iz);
	l0 = L(ix,nys - 1,iz);
	xp[ln] = xp[l0];
	xrho[ln] = xrho[l0];
	xwterm[ln] = xwterm[l0];
	xvx[ln] = xvx[l0];
	xvy[ln] = xvy[l0];
	xvz[ln] = xvz[l0];
	xent[ln] = xent[l0];
	xrt[ln] = xrt[l0];
	xrr[ln] = xrr[l0];
	xth[ln] = xth[l0];
      }
    }
  }
  else {   /* parallel processing */

    /* transfer needed stuff to buffer */
    position = 0;
    for (ix = -1; ix <= nxs; ix++) {
      for (iz = 0; iz < nz; iz++) {
	l0 = L(ix,nys - 1,iz);
	bbuffiny[position++] = xp[l0];
	bbuffiny[position++] = xrho[l0];
	bbuffiny[position++] = xwterm[l0];
	bbuffiny[position++] = xvx[l0];
	bbuffiny[position++] = xvy[l0];
	bbuffiny[position++] = xvz[l0];
	bbuffiny[position++] = xent[l0];
	bbuffiny[position++] = xrt[l0];
	bbuffiny[position++] = xrr[l0];
	bbuffiny[position++] = xth[l0];
      }
    }
    if (position > bsizey) {
      fprintf(stderr,"timestep: bbuffiny buffer not big enough\n");
      exit(54);
    }

    /* transfer stuff in positive y direction */
    transfer(bbuffiny,bbuffouty,rankb,rankt,position,oddigy,mycom);

    /* extract stuff from buffer */
    position = 0;
    for (ix = -1; ix <= nxs; ix++) {
      for (iz = 0; iz < nz; iz++) {
	ln = L(ix,-1,iz);
	xp[ln] = bbuffouty[position++];
	xrho[ln] = bbuffouty[position++];
	xwterm[ln] = bbuffouty[position++];
	xvx[ln] = bbuffouty[position++];
	xvy[ln] = bbuffouty[position++];
	xvz[ln] = bbuffouty[position++];
	xent[ln] = bbuffouty[position++];
	xrt[ln] = bbuffouty[position++];
	xrr[ln] = bbuffouty[position++];
	xth[ln] = bbuffouty[position++];
      }
    }
  }

  /* compute the centered intermediate fluxes -- no eddy mixing */
  for (ix = -1; ix <= nxs; ix++) {
    for (iy = -1; iy <= nys; iy++) {
      for (iz = 0; iz < nz; iz++) {
	l = L(ix,iy,iz);
	xthe[l] = thecomp(xent[l]);
	xrent[l] = xrho[l]*xent[l];
	xrrr[l] = xrho[l]*xrr[l];
	xrrt[l] = xrho[l]*xrt[l];
	fru[l] = xrho[l]*xvx[l];
	frv[l] = xrho[l]*xvy[l];
	frw[l] = xrho[l]*xvz[l];
	fruu[l] = fru[l]*xvx[l];
	fruv[l] = fru[l]*xvy[l];
	fruw[l] = fru[l]*xvz[l];
	frvv[l] = frv[l]*xvy[l];
	frvw[l] = frv[l]*xvz[l];
	frww[l] = frw[l]*xvz[l];
	fruent[l] = fru[l]*xent[l];
	frvent[l] = frv[l]*xent[l];
	frwent[l] = frw[l]*xent[l];
	frurt[l] = fru[l]*xrt[l];
	frvrt[l] = frv[l]*xrt[l];
	frwrt[l] = frw[l]*xrt[l];
	frurr[l] = fru[l]*xrr[l];
	frvrr[l] = frv[l]*xrr[l];
	frwrr[l] = (frw[l] - xrho[l]*xwterm[l])*xrr[l];
      }
    }
  }

  /* do half-step stuff -- fluxes include eddy mixing */
  for (ix = -1; ix < nxs; ix++) {
    for (iy = -1; iy < nys; iy++) {
      for (iz = 0; iz < nz - 1; iz++) {
	j = J(ix + 1,iy + 1,iz + 1);
	crho = CU(xrho,ix,iy,iz);
	swterm[j] = CU(xwterm,ix,iy,iz);

	/* compute half step values on grid corners */
	srho[j] = crho - FDIV(fru,frv,frw)*dthalf;
	sfru[j] = CU(fru,ix,iy,iz) - (FDIV(fruu,fruv,fruw) +
	   (SQYZ(xp,ix + 1,iy,iz) - SQYZ(xp,ix,iy,iz))/dx)*dthalf;
	sfrv[j] = CU(frv,ix,iy,iz) - (FDIV(fruv,frvv,frvw) +
	   (SQXZ(xp,ix,iy + 1,iz) - SQXZ(xp,ix,iy,iz))/dy)*dthalf;
	sfrw[j] = CU(frw,ix,iy,iz) - (FDIV(fruw,frvw,frww) +
	   (SQXY(xp,ix,iy,iz + 1) - SQXY(xp,ix,iy,iz))/dz + GEE*crho)*dthalf;
	srent[j] = CU(xrent,ix,iy,iz) - FDIV(fruent,frvent,frwent)*dthalf;
	srrt[j] = CU(xrrt,ix,iy,iz) - FDIV(frurt,frvrt,frwrt)*dthalf;
	srrr[j] = CU(xrrr,ix,iy,iz) - FDIV(frurr,frvrr,frwrr)*dthalf;
	sp[j] = CU(xp,ix,iy,iz);

	/* compute the bulk fluxes */
	sfruu[j] = sfru[j]*sfru[j]/srho[j];
	sfruv[j] = sfru[j]*sfrv[j]/srho[j];
	sfruw[j] = sfru[j]*sfrw[j]/srho[j];
	sfrvv[j] = sfrv[j]*sfrv[j]/srho[j];
	sfrvw[j] = sfrv[j]*sfrw[j]/srho[j];
	sfrww[j] = sfrw[j]*sfrw[j]/srho[j];
	sfruent[j] = srent[j]*sfru[j]/srho[j];
	sfrvent[j] = srent[j]*sfrv[j]/srho[j];
	sfrwent[j] = srent[j]*sfrw[j]/srho[j];
	sfrurt[j] = srrt[j]*sfru[j]/srho[j];
	sfrvrt[j] = srrt[j]*sfrv[j]/srho[j];
	sfrwrt[j] = srrt[j]*sfrw[j]/srho[j];
	sfrurr[j] = srrr[j]*sfru[j]/srho[j];
	sfrvrr[j] = srrr[j]*sfrv[j]/srho[j];
	sfrwrr[j] = srrr[j]*(sfrw[j]/srho[j] - swterm[j]);

	/* compute the eddy mixing coefficient */
	dxx = DX(xvx,ix,iy,iz);
	dyy = DY(xvy,ix,iy,iz);
	dzz = DZ(xvz,ix,iy,iz);
	dxy = 0.5*(DX(xvy,ix,iy,iz) + DY(xvx,ix,iy,iz));
	dyz = 0.5*(DY(xvz,ix,iy,iz) + DZ(xvy,ix,iy,iz));
	dzx = 0.5*(DZ(xvx,ix,iy,iz) + DX(xvz,ix,iy,iz));
	dsq = dxx*dxx + dyy*dyy + dzz*dzz +
	  2.*(dxy*dxy + dyz*dyz + dzx*dzx);
	rh = SQXY(xrt,ix,iy,iz)/rs(SQXY(xth,ix,iy,iz),
				   exner(SQXY(xp,ix,iy,iz)));
	mstab = 0.5*((rh - 1.)/drh + 1.);
	if (mstab > 1.) mstab = 1.;
	if (mstab < 0.) mstab = 0.;
	dstab = 1. - mstab;
	zgrad = dstab*DZ(xth,ix,iy,iz) + mstab*DZ(xthe,ix,iy,iz);
	bvsq = GEE*zgrad/SQXY(xth,ix,iy,iz);
	kcriterion = dsq - 2.*bvsq;
	if (kcriterion < 0.) kcriterion = 0.;
	skk[j] = eddy*crho*pow(kcriterion,0.5)*dz*dz;
	skv[j] = skk[j];

	/* add on eddy diffusion fluxes if K > 0 */
	if (skv[j] > 0. ) {
	  sfruu[j] -= skv[j]*dxx;
	  sfrvv[j] -= skv[j]*dyy;
	  sfrww[j] -= skv[j]*dzz;
	  sfruv[j] -= skv[j]*dxy;
	  sfruw[j] -= skv[j]*dzx;
	  sfrvw[j] -= skv[j]*dyz;
	}
	if (skk[j] > 0. ) {
	  sfruent[j] -= skk[j]*DX(xent,ix,iy,iz);
	  sfrvent[j] -= skk[j]*DY(xent,ix,iy,iz);
	  sfrwent[j] -= skk[j]*DZ(xent,ix,iy,iz);
	  sfrurt[j] -= skk[j]*DX(xrt,ix,iy,iz);
	  sfrvrt[j] -= skk[j]*DY(xrt,ix,iy,iz);
	  sfrwrt[j] -= skk[j]*DZ(xrt,ix,iy,iz);
	  sfrurr[j] -= skk[j]*DX(xrr,ix,iy,iz);
	  sfrvrr[j] -= skk[j]*DY(xrr,ix,iy,iz);
	}

	/* special case of extra smoothing for liquid water */
	sfrwrr[j] -= (skk[j] + crho*lambda*dz*dz/dt)*DZ(xrr,ix,iy,iz);
      }
    }
  }

  /* mirror fluxes across the rigid upper and lower boundaries */
  for (ix = 0; ix <= nxs; ix++) {
    for (iy = 0; iy <= nys; iy++) {
      j0 = J(ix,iy,0);
      j1 = J(ix,iy,1);
      j2 = J(ix,iy,2);
      jl = J(ix,iy,nz - 2);
      jm = J(ix,iy,nz - 1);
      jn = J(ix,iy,nz);
      sfruu[j0] = sfruu[j1];
      sfruu[jn] = sfruu[jm];
      sfruv[j0] = sfruv[j1];
      sfruv[jn] = sfruv[jm];
      sfruw[j0] = -sfruw[j1];
      sfruw[jn] = -sfruw[jm];;
      sfrvv[j0] = sfrvv[j1];
      sfrvv[jn] = sfrvv[jm];
      sfrvw[j0] = -sfrvw[j1];
      sfrvw[jn] = -sfrvw[jm];
      sfrww[j0] = sfrww[j1];
      sfrww[jn] = sfrww[jm];
      sfru[j0] = sfru[j1];
      sfru[jn] = sfru[jm];
      sfrv[j0] = sfrv[j1];
      sfrv[jn] = sfrv[jm];
      sfrw[j0] = -sfrw[j1];
      sfrw[jn] = -sfrw[jm];
      sfruent[j0] = sfruent[j1];
      sfruent[jn] = sfruent[jm];
      sfrvent[j0] = sfrvent[j1];
      sfrvent[jn] = sfrvent[jm];
      sfrwent[j0] = -sfrwent[j1];
      sfrwent[jn] = -sfrwent[jm];
      sfrurt[j0] = sfrurt[j1];
      sfrurt[jn] = sfrurt[jm];
      sfrvrt[j0] = sfrvrt[j1];
      sfrvrt[jn] = sfrvrt[jm];
      sfrwrt[j0] = -sfrwrt[j1];
      sfrwrt[jn] = -sfrwrt[jm];
      sfrurr[j0] = sfrurr[j1];
      sfrurr[jn] = sfrurr[jm];
      sfrvrr[j0] = sfrvrr[j1];
      sfrvrr[jn] = sfrvrr[jm];
      sfrwrr[j0] = 2.*sfrwrr[j1] - sfrwrr[j2];
      sfrwrr[jn] = 0.;
      srho[j0] = 2.*srho[j1] - srho[j2];
      srho[jn] = 2.*srho[jm] - srho[jl];
      srent[j0] = 2.*srent[j1] - srent[j2];
      srent[jn] = 2.*srent[jm] - srent[jl];
      srrt[j0] = 2.*srrt[j1] - srrt[j2];
      srrt[jn] = 2.*srrt[jm] - srrt[jl];
      srrr[j0] = 2.*srrr[j1] - srrr[j2];
      srrr[jn] = 2.*srrr[jm] - srrr[jl];
      sp[j0] = 2.*sp[j1] - sp[j2];
      sp[jn] = 2.*sp[jm] - sp[jl];
      skk[j0] = skk[j1];
      skk[jn] = skk[jm];
      skv[j0] = skv[j1];
      skv[jn] = skv[jm];
    }
  }

  /* calculate potential temperature and pressure on grid corners */
  for (ix = 0; ix <= nxs; ix++) {
    for (iy = 0; iy <= nys; iy++) {
      for (iz = 0; iz <= nz; iz++) {
	j = J(ix,iy,iz);
	sth = thetacomp2(srent[j]/srho[j],srrt[j]/srho[j],exner(sp[j]));
	sp[j] = P_REF*pow(srho[j]*R*sth/P_REF,GAMMA);
      }
    }
  }

  /* compute full step values -- extract selected vertical fluxes */
  for (ix = 0; ix < nxs; ix++) {
    for (iy = 0; iy < nys; iy++) {
      for (iz = 0; iz < nz; iz++) {
	i = I(ix,iy,iz);

	/* density */
	oldrho = rho[i];
	rho[i] = rho[i] - (FDIV2(sfru,sfrv,sfrw) -
			   DFILT(xrho,ix,iy,iz)*filter)*dt;

	/* x velocity */
	vx[i] = vx[i]*oldrho -
	  (FDIV2(sfruu,sfruv,sfruw) - DFILT(fru,ix,iy,iz)*filter +
	   (SQ2YZ(sp,ix + 1,iy,iz) - SQ2YZ(sp,ix,iy,iz))/dx)*dt -
	  rho[i]*(svx[iz] + mu[iz]*(vx[i] - vx0[iz]))*dt;

	/* y velocity */
	vy[i] = vy[i]*oldrho -
	  (FDIV2(sfruv,sfrvv,sfrvw) - DFILT(frv,ix,iy,iz)*filter +
	   (SQ2XZ(sp,ix,iy + 1,iz) - SQ2XZ(sp,ix,iy,iz))/dy)*dt -
	  rho[i]*(svy[iz] + mu[iz]*(vy[i] - vy0[iz]))*dt;

	/* z velocity */
	if (iz == 0 || iz == nz - 1) {
	  vz[i] = 0.;
	}
	else {
	  crho = 0.5*(SQ2XY(srho,ix,iy,iz + 1) + SQ2XY(srho,ix,iy,iz));
	  vz[i] = vz[i]*oldrho - (FDIV2(sfruw,sfrvw,sfrww) -
				    DFILT(frw,ix,iy,iz)*filter +
	     (SQ2XY(sp,ix,iy,iz + 1) - SQ2XY(sp,ix,iy,iz))/dz +
	     GEE*crho)*dt - rho[i]*mu[iz]*vz[i]*dt;
	}

	/* specific moist entropy */
	ent[i] = ent[i]*oldrho -
	  (FDIV2(sfruent,sfrvent,sfrwent) - DFILT(xrent,ix,iy,iz)*filter)*dt +
	  rho[i]*(sentrad[i] - sentwtg[iz])*dt;

	/* total cloud water mixing ratio */
	rt[i] = rt[i]*oldrho -
	  (FDIV2(sfrurt,sfrvrt,sfrwrt) -
	   DFILT(xrrt,ix,iy,iz)*filter)*dt - rho[i]*srtwtg[iz]*dt;

	/* precipitation mixing ratio */
	rr[i] = rr[i]*oldrho - (FDIV2(sfrurr,sfrvrr,sfrwrr) -
				  DFILT(xrrr,ix,iy,iz)*filter)*dt;
	
	/* at this point vars are still multiplied by rho -- divide out */
	vx[i] /= rho[i];
	vy[i] /= rho[i];
	vz[i] /= rho[i];
	ent[i] /= rho[i];
	rt[i] /= rho[i];
	rr[i] /= rho[i];

	/* do precipitation processes only on full step */
	temperature = th[i]*pow(p[i]/P_REF,KAPPA);
	rsat = rs(th[i],exner(p[i]));
	rc[i] = rt[i] > rsat ? rt[i] - rsat : 0.;
	r[i] = rt[i] - rc[i];
	cprecip = temperature > FREEZING ? crain : csnow;
	newrain = (cprecip*rc[i] - cevap*(rsat - r[i])*rr[i])*dt;
	if (newrain > rc[i]) newrain = rc[i];
	if (newrain < -rr[i]) newrain = -rr[i];
	rr[i] += newrain;
	rt[i] -= newrain;
	srtrain[i] = -newrain/dt;
	rc[i] = rt[i] > rsat ? rt[i] - rsat : 0.;
	r[i] = rt[i] - rc[i];

	/* compute cumulus fluxes on I grid */
	vxcuflux[i] = CU2(sfruw,ix,iy,iz);
	vycuflux[i] = CU2(sfrvw,ix,iy,iz);
	entcuflux[i] = CU2(sfrwent,ix,iy,iz);
	rtcuflux[i] = CU2(sfrwrt,ix,iy,iz);
	rrcuflux[i] = CU2(sfrwrr,ix,iy,iz);

	/* surface stuff */
	if (iz == 0) {
	  k = K(ix,iy);

	  /* update the surface cumulus fluxes with old surface values
	   * to match what is done with convective fluxes aloft
	   */
	  vxcuflux[i] = vxflux[k];
	  vycuflux[i] = vyflux[k];
	  entcuflux[i] = entflux[k];
	  rtcuflux[i] = rtflux[k];

	  /* recompute surface fluxes */
	  ueff = pow(vx[i]*vx[i] + vy[i]*vy[i] + ueffmin*ueffmin,0.5);
	  pisurf = exner(p[i]);
	  thsurf = stemp[k]*CP/pisurf;
	  if (thsurf > th[i]) {
	    entsurf = entscomp(thsurf,pisurf);
	    rtsurf = rs(thsurf,pisurf);
	    entflux1 = cdrag*ueff*(entsurf - ent[i]);
	    rtflux1 = cdrag*ueff*(rtsurf - rt[i]);
	    vxflux1 = -cdrag*ueff*vx[i];
	    vyflux1 = -cdrag*ueff*vy[i];
	  }
	  else {
	    entflux1 = rtflux1 = vxflux1 = vyflux1 = 0.;
	  }
	  ent[i] += dt*entflux1/dzhalf;
	  rt[i] += dt*rtflux1/dzhalf;
	  vx[i] += dt*vxflux1/dzhalf;
	  vy[i] += dt*vyflux1/dzhalf;
	  entflux[k] = rho[i]*entflux1;
	  entfluxcum[k] += entflux[k]*dt;
	  rtflux[k] = rho[i]*rtflux1;
	  rtfluxcum[k] += rtflux[k]*dt;
	  vxflux[k] = rho[i]*vxflux1;
	  vxfluxcum[k] += vxflux[k]*dt;
	  vyflux[k] = rho[i]*vyflux1;
	  vyfluxcum[k] += vyflux[k]*dt;
	  rainflux[k] = SQ2XY(sfrwrr,ix,iy,0);
	  rainfluxcum[k] += rainflux[k]*dt;
	}

	/* accumulate the convective fluxes */
	vxcufluxcum[i] += vxcuflux[i]*dt;
	vycufluxcum[i] += vycuflux[i]*dt;
	entcufluxcum[i] += entcuflux[i]*dt;
	rtcufluxcum[i] += rtcuflux[i]*dt;
	rrcufluxcum[i] += rrcuflux[i]*dt;

	/* this is for information only */
	kk[i] = CU2(skk,ix,iy,iz);
	kv[i] = CU2(skv,ix,iy,iz);
      }

      /* redistribute rt to eliminate negative values, which only
	 occur near the top of the domain -- borrow from below */
      for (iz = nz - 1; iz >= 0; iz--) {
	i = I(ix,iy,iz);
	if (rt[i] < 0. && iz > 0) {
	  ilo = I(ix,iy,iz - 1);
	  rt[ilo] += rt[i]*rho[i]/rho[ilo];
	  rt[i] = 0.;
	}

	/* compute potential temp and pressure */
	th[i] = thetacomp2(ent[i],rt[i],exner(p[i]));
	p[i] = P_REF*pow(rho[i]*R*th[i]/P_REF,GAMMA);

	/* accumulate sentrad and srtrain */
	sentradcum[i] += sentrad[i]*dt;
	srtraincum[i] += srtrain[i]*dt;
      }
    }
  }
}
