/* toyrad.c -- Toy radiation parameterization from diabat2.c.
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "toyrad.h"

/* constants for radiation */
#define TREF 300.0
#define RHOREF 1.2
#define CP 1005.0
#define CO2 0
#define CONT 1
#define RTREF 0.02
#define SBCONST 5.67e-8
#define SOLARCONST 1360.

/* safeexp -- Do our own computation of the exponential function --
 * this results from the sporadic long times required to compute
 * the exponential function using the standard GNU math library on
 * 64 bit systems (AMD, Intel)
 */
#define E 2.7182818284590451
#define ETABLESIZE 100001
static double safeexp(double x)
{
  static double xx, result, table[ETABLESIZE];
  static int begin = 1, i;

  if (begin) {
    for (i = 0; i < ETABLESIZE; i++) {
      table[i] = exp(((double)i)/((double)(ETABLESIZE)));
    }
    begin = 0;
  }

  xx = x;
  result = 1.;
  while (xx >= 0.) {
    xx -= 1.;
    result *= E;
  }
  while (xx < 0.) {
    xx += 1.;
    result /= E;
  }
  i = xx*ETABLESIZE;
  if (i < 0) i = 0;
  if (i >= ETABLESIZE) i = ETABLESIZE - 1;
  result *= table[i];
  return(result);
}

/* floatbuff -- This routine uses malloc to allocate float buffer
 * space.  The argument nelem defines the number of floating
 * point elements.  It returns a float pointer to the space.  If malloc
 * fails, getbuff dies with an error message.
 */
static double *floatbuff(int nelem)
{
  double *point;
  if ((point = (double*)malloc((unsigned)(sizeof(double)*nelem)))
       == (double*)NULL) {
    fprintf(stderr,"floatbuff: can't get %d words\n",nelem);
    exit(1);
  }
  return(point);
}




/* radiation -- Parameterization of long and shortwave radiation
 * this model is tuned by making rough comparisons with NCAR's
 * CCM2 radiation model (constructed by Jeff Kiehl and associates)
 */

/* one and only one of the following radiation options must be turned
   on (on = 1, off = 0) -- 1 channel radiation is faster, but is just
   a gray body scheme -- 8 channel radiation is the radiation scheme
   used in the Raymond-Torres (1998) paper -- J. Atmos. Sci., 55,
   1771-1790 -- and includes a crude representation of radiation-water
   vapor interactions */
#define ONECHANNEL 0
#define MULTICHANNEL 0
#define MEGACHANNEL 1

/* constants for the radiation parameterization */
#if ONECHANNEL
#define NBANDS 1
#endif

#if MULTICHANNEL
#define NBANDS 8
#endif

#if MEGACHANNEL
#define NBANDS 13
#endif

/* elimit -- limit the value to +- 75 to avoid blowing up exponential fn */
static double elimit(val)
double val;
{
  double lval;
  lval = val;
  if (lval > 75.) lval = 75.;
  if (lval < -75.) lval = -75.;
  return(lval);
}

/* lcomp -- compute the l factor from delta tau and the b values
 * this assumes that b = sigma*T^4 is linear across the grid cell
 * -- integrate analytically and watch out for small dtau
 */
static double lcomp(double dtau, double b1, double b2)
{
  double efact,omefact,lfactor;
  efact = safeexp(elimit(-dtau));
  omefact = 1. - efact;
  if (dtau < 0.001) {
    lfactor = 0.5*dtau*(b1 + b2);
  }
  else {
    lfactor = b1*(omefact/dtau - efact) + b2*(1. - omefact/dtau);
  }
  return(lfactor);
}

/* kcomp -- compute the kplus and kminus profiles */
static void kcomp(long nz, double *tau, double *th,
		   double *pi, double *kplus, double *kminus)
{
  long iz,jz;
  double temp,dtau,efact;
  static double *b;                  /* extrapolated cell edge sigma*T^4 */
  static int begin = 1;

  /* make space only once */
  if (begin) {
    b = floatbuff(nz);
    begin = 0;
  }

  /* compute the Planck function profile on cell edges */
  for (iz = 0; iz < nz; iz++) {
    temp = th[iz]*pi[iz]/CP;
    b[iz] = SBCONST*temp*temp*temp*temp;
  }

  /* do the kplus profile */
  kplus[0] = 0.;
  for (iz = 1; iz < nz; iz++) {
    jz = iz - 1;
    dtau = tau[iz] - tau[jz];
    efact = safeexp(elimit(-dtau));
    kplus[iz] = efact*kplus[jz] + lcomp(dtau, b[jz], b[iz]);
    if (fabs(kplus[iz]) < 1.e-30) kplus[iz] = 0.;
  }

  /* do the kminus profile */
  kminus[nz - 1] = 0.;
  for (iz = nz - 2; iz >= 0; iz--) {
    jz = iz + 1;
    dtau = tau[jz] - tau[iz];
    efact = safeexp(elimit(-dtau));
    kminus[iz] = efact*kminus[jz] + lcomp(dtau,b[jz],b[iz]);
    if (fabs(kminus[iz]) < 1.e-30) kminus[iz] = 0.;
  }
}





/* taucomp -- compute the path length for a band -- tau is defined on
 * cell edges and is zero at the surface
 */
static void taucomp(long nz, long ib, double *th, double *zc,
		    double *rt, double *pi, double *rho, double *rsat,
		    double abs, double cld, double *tau)
{
  long iz,kz;
  double mutot,temp,pbroad,rl,rvap,mutotmax,rfract;

  /* integrate upward from surface to get tau -- iz = upper edge index,
   * kz = lower edge index
   */
  tau[0] = 0.;
  for (iz = 1; iz < nz; iz++) {
    kz = iz - 1;
    
    /* compute pressure broadening factor */
    temp = 0.5*(th[iz]*pi[iz] + th[kz]*pi[kz])/CP;
    pbroad = pow(temp/TREF,0.5)*0.5*(rho[iz] + rho[kz])/RHOREF;

    /* compute vapor-condensate split */
    rl = 0.5*(rt[iz] - rsat[iz] + rt[kz] - rsat[kz]);
    if (rl < 0.) rl = 0.;
    rvap = 0.5*(rt[iz] + rt[kz]) - rl;
    rfract = rvap/RTREF;

    /* compute mutot in cell center at level kz for desired band */
    if (ib == CO2) {                 /* carbon dioxide bands */
      mutot = abs*pbroad;
    }
    else if (ib == CONT) {           /* water vapor continuum */
      mutot = abs*rfract*rfract;
    }
    else {                           /* water vapor bands */
      mutot = abs*pbroad*rfract;
    }

    /* add in cloud part */
    mutot += cld*rl;

    /* compute tau at the next level */
    tau[iz] = tau[kz] + 0.5*(rho[iz] + rho[kz])*(zc[iz] - zc[kz])*mutot;
  }
}





/* the main radiation routine */
void radiation(long nz, long nosun, double cld, double stemp,
	       double albedo, double *zc, double *th,
	       double *ent, double *rt, double *pi, double *rho,
	       double *rsat, double *sent, double *flux)
{
  long ib,iz;                 /* index parameters */
  double scos;                /* factor for solar zenith angle */
  double bsurf;               /* upward energy flux from surface */
  double topefact,botefact,iplus,iminus;  /* intermediate results */
  double normfactor;          /* normalization factor */
  double visible;             /* visible solar radiation reaching surface */
  static double *eflux;              /* total flux on cell edges */
  static double *tau;                /* actual absorptivities */
  static double *kplus,*kminus;      /* intermediate profiles */
  static int begin = 1;

  /* arrays containing per-band information about radiation model */

#if ONECHANNEL
  /* this has a single band independent of water vapor */

  /* specific absorption coefficients for each band */
  static double abs[NBANDS] = {1.};

  /* downward flux of solar radiation for each band for overhead sun */
  static double solar[NBANDS] = {0.};

  /* fraction of Planck-weighted spectrum occupied by each band */
  static double fract[NBANDS] = {1.0};
#endif

#if MULTICHANNEL
  /* arrays containing per-band information about radiation model band
   * 0 is CO2 while band 1 is water vapor continuum bands 2-7 are
   * water vapor lines.  This is the radiation scheme used in Raymond
   * and Torres (1998)
   */

  /* specific absorption coefficients for each band */
  static float abs[NBANDS] = {2.,0.5,1.,10.,100.,1000.,10000.,100000.};

  /* downward flux of solar radiation for each band for overhead sun */
  static float solar[NBANDS] = {0.,160.,60.,30.,15.,7.,0.,0.};

  /* fraction of Planck-weighted spectrum occupied by each band */
  static float fract[NBANDS] = {0.11,0.20,0.19,0.16,0.13,0.10,0.07,0.04};
#endif

#if MEGACHANNEL
  /* arrays containing per-band information about radiation model band
   * 0 is CO2 while band 1 is water vapor continuum bands 2-12 are
   * water vapor lines.
   */

  /* specific absorption coefficients for each band */
  static float abs[NBANDS] = {2.,0.5,1.,3.16,10.,31.6,100.,316.,
			      1000.,3160.,10000.,31600.,100000.};

  /* downward flux of solar radiation for each band for overhead sun */
  static float solar[NBANDS] = {0.,160.,35.,26.,19.,13.,10.,6.,3.,
				0.,0.,0.,0.};

  /* fraction of Planck-weighted spectrum occupied by each band */
  static float fract[NBANDS] = {0.11,0.20,0.13,0.11,0.09,0.08,0.07,0.06,
				0.05,0.04,0.03,0.02,0.01};
#endif

  /* make space */
  if (begin) {
    eflux = floatbuff(nz);
    tau = floatbuff(nz);
    kplus = floatbuff(nz);
    kminus = floatbuff(nz);
    begin = 0;
  }

  /* define solar zenith angle factor and surface Planck function
   * we include the nosun flag for testing the scheme with no
   * solar input from the top
   */
  if (nosun) {
    scos = 0.;
  }
  else {
    scos = 1./3.14159;
  }
  bsurf = SBCONST*stemp*stemp*stemp*stemp;

  /* zero edge flux array */
  for (iz = 0; iz < nz; iz++) eflux[iz] = 0.;

  /* loop on bands */
  for (ib = 0; ib < NBANDS; ib++) {

    /* compute tau, the actual absorption coefficient, for this band */
    taucomp(nz,ib,th,zc,rt,pi,rho,rsat,abs[ib],cld,tau);

    /* compute the kplus and kminus profiles */
    kcomp(nz,tau,th,pi,kplus,kminus);

    /* add in the flux profile for this band */
    for (iz = 0; iz < nz; iz++) {
      botefact = safeexp(elimit(-tau[iz]));
      topefact = safeexp(elimit(tau[iz] - tau[nz - 1]));
      iplus = fract[ib]*(bsurf*botefact + kplus[iz]);
      iminus = scos*solar[ib]*topefact + fract[ib]*kminus[iz];
      eflux[iz] += (iplus - iminus);
    }
  }

  /* compute solar radiation reaching surface assuming zenith sun */
  visible = SOLARCONST*(1. - albedo);
  for (ib = 0; ib < NBANDS; ib++) {
    visible -= solar[ib];
  }

  /* compute the cell-edged radiative source and flux */
  for (iz = 0; iz < nz; iz++) {

    /* convert from energy/vol/time to entropy/mass/time */
    normfactor = CP/(rho[iz]*pi[iz]*th[iz]);

    /* compute the flux divergence to get entropy source from IR */
    if (iz == 0) {
      sent[iz] = -normfactor*(eflux[1] - eflux[0])/(zc[1] - zc[0]);
    }
    else if (iz == nz - 1) {
      sent[iz] = -normfactor*(eflux[iz] - eflux[iz - 1])/(zc[iz] - zc[iz - 1]);
    }
    else {
      sent[iz] = -normfactor*(eflux[iz + 1] - eflux[iz - 1])/
	(zc[iz + 1] - zc[iz - 1]);
    }

    /* combine IR and visible radiation fluxes */
    flux[iz] = eflux[iz] - scos*visible;
  }
}





/* fixed radiation code */
void fradiation(long nz, double *zc, double *pi, double *th,
		double radcool, double tpause, double radbrk,
		double *ent, double *sentrad, double *sentradclr,
		double *irflux, double *irfluxclr)
{
  long iz;
  double z, radentsrc, temperature;

  for (iz = 0; iz < nz; iz++) {
    z = zc[iz];

    /* this converts a temperature cooling rate to an entropy source */
    temperature = pi[iz]*th[iz]/CP;
    radentsrc = -CP*radcool/temperature;

    /* compute the fixed radiative cooling temperature tendency profile */
    if (z <= radbrk*tpause) {
      sentrad[iz] = 1.;
    }
    else if (z < tpause && z > radbrk*tpause) {
      sentrad[iz] = 1. - (z/tpause - radbrk)/(1. - radbrk);
    }
    else {
      sentrad[iz] = 0.;
    }

    /* this converts to units of joules/kilo/kelvin/kilosecond */
    sentrad[iz] = sentrad[iz]*radentsrc/86.4;

    /* fill in other stuff -- ignore irflux var for now */
    sentradclr[iz] = sentrad[iz];
    irflux[iz] = irfluxclr[iz] = 0.;
  }
}
